__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import os
from contextlib import contextmanager
from typing import Literal

import torch
from pydantic import BaseModel
from torch.distributed import destroy_process_group, init_process_group


class ElasticRunVars(BaseModel):
    local_rank: int
    world_size: int
    is_cuda: bool


def _torchrun_env_variables(is_cuda: bool) -> ElasticRunVars:
    """
    Retrieve environment variables set by torchrun in a distributed setting.
    See https://pytorch.org/docs/stable/elastic/run.html#environment-variables.
    """
    local_rank = os.environ.get("LOCAL_RANK")
    world_size = os.environ.get("WORLD_SIZE")
    assert local_rank and world_size
    return ElasticRunVars(local_rank=int(local_rank), world_size=int(world_size), is_cuda=is_cuda)


@contextmanager
def process_group(*, backend: Literal["gloo", "nccl"] | None = None):
    """
    Context manager for initializing a distributed process group and
    automatically destroying resources upon exit. See
    https://pytorch.org/docs/stable/distributed.html

    Args:
        backend ("gloo", "nccl" or None = None): By default, uses nccl if CUDA is
            available, else gloo. Set this explicitly if you want to use nccl in
            the presence of CUDA-incompatible GPUs

    **Example**::

        with process_group() as run_vars:
            trainer = MyTrainer(
                config,
                gpu_id=run_vars.local_rank,
            ).trainer.train()

    """
    if not backend:
        backend = "nccl" if torch.cuda.is_available() else "gloo"

    init_process_group(backend=backend)
    is_cuda = backend == "nccl"
    run_vars = _torchrun_env_variables(is_cuda)

    if backend == "nccl":
        torch.cuda.set_device(run_vars.local_rank)
    try:
        yield run_vars
    finally:
        destroy_process_group()



"""Package initialization."""

__copyright__ = """
LICENSED INTERNAL CODE. PROPERTY OF IBM.
IBM Research Licensed Internal Code
(C) Copyright IBM Corp. 2024
ALL RIGHTS RESERVED
"""

__version__ = "0.0.1"


from mblm.model.config import MBLMModelConfig, MBLMReturnType
from mblm.model.mamba import MambaBlockConfig
from mblm.model.mblm import MBLM
from mblm.model.transformer import TransformerBlockConfig

__all__ = [
    "MBLM",
    "MBLMModelConfig",
    "MBLMReturnType",
    "TransformerBlockConfig",
    "MambaBlockConfig",
]



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import codecs
from typing import TextIO

import torch


def log(t: torch.Tensor, eps: float = 1e-20):
    return torch.log(t.clamp(min=eps))


def gumbel_noise(t: torch.Tensor):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))


def gumbel_sample(logits: torch.Tensor, temperature: float = 1.0, dim: int = -1):
    return ((logits / temperature) + gumbel_noise(logits)).argmax(dim=dim)


def top_k(logits: torch.Tensor, thres: float = 0.5):
    num_logits = logits.shape[-1]
    k = max(int((1 - thres) * num_logits), 1)
    val, ind = torch.topk(logits, k)
    probs = torch.full_like(logits, float("-inf"))
    probs.scatter_(1, ind, val)
    return probs


class RoPE(torch.nn.Module):
    """
    Inlined and with adjusted cache size for dynamic sequence lengths from
    https://pytorch.org/torchtune/0.2/_modules/torchtune/modules/position_embeddings.html#RotaryPositionalEmbeddings:

    This class implements Rotary Positional Embeddings (RoPE) proposed in
    https://arxiv.org/abs/2104.09864.

    Reference implementation (used for correctness verfication) can be found
    here: https://github.com/meta-llama/llama/blob/main/llama/model.py#L80

    In this implementation we cache the embeddings for each position upto
    ``max_seq_len`` by computing this during init.

    Args:
        dim (int): Embedding dimension. This is usually set to the dim of each
            head in the attention module computed as ````embed_dim`` //
            ``num_heads````
        max_seq_len (int): Maximum expected sequence length for the
            model, if exceeded the cached freqs will be recomputed
        base (int): The base for the geometric progression used to compute
            the rotation angles
    """

    cache: torch.Tensor
    theta: torch.Tensor

    def __init__(
        self,
        dim: int,
        max_seq_len: int,
        base: int = 10_000,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.base = base
        self.max_seq_len = max_seq_len
        self._rope_init()

    # We need to explicitly define reset_parameters for FSDP initialization, see
    # https://github.com/pytorch/pytorch/blob/797d4fbdf423dd9320ebe383fb57ffb1135c4a99/torch/distributed/fsdp/_init_utils.py#L885
    def reset_parameters(self):
        self._rope_init()

    def _rope_init(self):
        theta = 1.0 / (
            self.base ** (torch.arange(0, self.dim, 2)[: (self.dim // 2)].float() / self.dim)
        )
        self.register_buffer("theta", theta, persistent=False)
        self.build_rope_cache(self.max_seq_len)

    def build_rope_cache(self, max_seq_len) -> None:
        # Create position indexes `[0, 1, ..., max_seq_len - 1]`
        seq_idx = torch.arange(max_seq_len, dtype=self.theta.dtype, device=self.theta.device)

        # Outer product of theta and position index; output tensor has a shape
        # of [max_seq_len, dim // 2]
        idx_theta = torch.einsum("i, j -> ij", seq_idx, self.theta).float()

        # cache includes both the cos and sin components and so the output shape
        # is [max_seq_len, dim // 2, 2]
        cache = torch.stack([torch.cos(idx_theta), torch.sin(idx_theta)], dim=-1)
        self.register_buffer("cache", cache, persistent=False)

    def forward(self, x: torch.Tensor, *, input_pos: torch.Tensor | None = None) -> torch.Tensor:
        """
        Args:
            x (Tensor): input tensor with shape
                [b, s, n_h, h_d]
            input_pos (Optional[Tensor]): Optional tensor which contains the position ids
                of each token. During training, this is used to indicate the positions
                of each token relative to its sample when packed, shape [b, s].
                During inference, this indicates the position of the current token.
                If none, assume the index of the token is its position id. Default is None.

        Returns:
            Tensor: output tensor with RoPE applied

        Notation used for tensor shapes:
            - b: batch size
            - s: sequence length
            - n_h: num heads
            - h_d: head dim

        TODO: The implementation below can be made more efficient
        for inference.
        """
        # input tensor has shape [b, s, n_h, h_d]
        seq_len = x.size(1)

        # for extrapolation scenarios, dynaimcally adjust the cache size
        if seq_len > self.max_seq_len:
            self.max_seq_len = seq_len
            self.build_rope_cache(seq_len)

        # extract the values based on whether input_pos is set or not
        rope_cache = self.cache[:seq_len] if input_pos is None else self.cache[input_pos]

        # reshape input; the last dimension is used for computing the output.
        # Cast to float to match the reference implementation
        # tensor has shape [b, s, n_h, h_d // 2, 2]
        xshaped = x.float().reshape(*x.shape[:-1], -1, 2)

        # reshape the cache for broadcasting
        # tensor has shape [b, s, 1, h_d // 2, 2] if packed samples,
        # otherwise has shape [1, s, 1, h_d // 2, 2]
        rope_cache = rope_cache.view(-1, xshaped.size(1), 1, xshaped.size(3), 2)

        # tensor has shape [b, s, n_h, h_d // 2, 2]
        x_out = torch.stack(
            [
                xshaped[..., 0] * rope_cache[..., 0] - xshaped[..., 1] * rope_cache[..., 1],
                xshaped[..., 1] * rope_cache[..., 0] + xshaped[..., 0] * rope_cache[..., 1],
            ],
            -1,
        )

        # tensor has shape [b, s, n_h, h_d]
        x_out = x_out.flatten(3)
        return x_out.type_as(x)


class ByteToUtf8Streamer:
    def __init__(self, stream: TextIO):
        self.stream = stream
        self.decoder = codecs.getincrementaldecoder("utf-8")()

    def write(self, single_byte: int):
        try:
            as_byte = single_byte.to_bytes(1, byteorder="big")
            decoded_str = self.decoder.decode(as_byte)

            self.stream.write(decoded_str)
            self.stream.flush()
        except UnicodeDecodeError:
            pass

    def flush_remaining(self):
        remaining = self.decoder.decode(b"", final=True)
        if remaining:
            self.stream.write(remaining)
            self.stream.flush()



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


from typing import Callable, Iterable, cast

import torch
from MEGABYTE_pytorch.megabyte import Attention, FeedForward, RMSNorm, RotaryEmbedding, token_shift
from pydantic import BaseModel, Field

from mblm.model.block import StageBlock


class TransformerBlockConfig(StageBlock, BaseModel):
    """
    General config for creating a Transformer Decocer block inside MBLM.
    """

    attn_head_dims: int
    attn_num_heads: int
    attn_use_rot_embs: bool
    attn_dropout: float = 0.0
    ff_multiplier: int = 2
    ff_dropout: float = 0.0
    use_flash_attn: bool = False

    block_type: str = Field(init=False, default="transformer")

    def to_model(self, model_dim, num_layers):
        return TransformerDecoder(
            model_dim=model_dim,
            num_layers=num_layers,
            attn_head_dim=self.attn_head_dims,
            attn_num_heads=self.attn_num_heads,
            attn_dropout=self.attn_dropout,
            attn_use_rot_embs=self.attn_use_rot_embs,
            ff_dropout=self.ff_dropout,
            ff_mult=self.ff_multiplier,
            use_flash_attn=self.use_flash_attn,
        )


class TransformerDecoder(torch.nn.Module):
    def __init__(
        self,
        *,
        model_dim: int,
        num_layers: int,
        attn_head_dim: int = 64,
        attn_num_heads: int = 8,
        attn_dropout: float = 0.0,
        attn_use_rot_embs: bool = True,
        ff_dropout: float = 0.0,
        ff_mult: int = 4,
        use_flash_attn: bool = False,
    ):
        super().__init__()
        self.rotary_emb = RotaryEmbedding(attn_head_dim) if attn_use_rot_embs else None

        self.layers = torch.nn.ModuleList([])

        for _ in range(num_layers):
            self.layers.append(
                torch.nn.ModuleList(
                    [
                        Attention(
                            dim=model_dim,
                            dim_head=attn_head_dim,
                            heads=attn_num_heads,
                            dropout=attn_dropout,
                            flash=use_flash_attn,
                        ),
                        FeedForward(dim=model_dim, mult=ff_mult, dropout=ff_dropout),
                    ]
                )
            )

        self.norm = RMSNorm(model_dim)

    def forward(self, input_ids: torch.Tensor) -> torch.Tensor:
        n = input_ids.shape[-2]
        rotary_emb: torch.Tensor | None = self.rotary_emb(n) if self.rotary_emb else None

        for attn, ff in cast(Iterable[tuple[Callable, Callable]], self.layers):
            input_ids = attn(token_shift(input_ids), rotary_emb=rotary_emb) + input_ids
            input_ids = ff(token_shift(input_ids)) + input_ids

        return self.norm(input_ids)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


from pydantic import BaseModel, Field

from mblm.model.block import StageBlock
from mblm.model.mamba_shim import Mamba1, Mamba1Config, Mamba2Mixer


class MambaBlockConfig(StageBlock, BaseModel):
    """
    General config for creating a Mamba block inside MBLM.
    Uses roughly 3 * expand * d_model^2 parameters.

    Parameters in brackets [x] denote the notation used in Mambabyte

    Parameters:
        n_layers: Number of layers [n]
        d_model: Residual stream dimension [d]
        expand: Linear layers expansion factor (always 2) [e]
        d_state: SSM state dimension [n_state]
        d_conv: Convolutional kernel size (always 4) [k]
        headdim: Head dimension
        dropout: The dropout in the Mamba block
    Note:
        `headdim` is a Mamba2 parameter only. There is no `dt_rank` [r]
        low-rank projection dimension in Mamba2
    """

    d_state: int
    d_conv: int
    expand: int
    headdim: int  # Mamba2 only
    dropout: float = 0.0  # Mamba2 only. Default for backwards compatibility

    block_type: str = Field(init=False, default="mamba1" if Mamba2Mixer is None else "mamba2")

    def to_model(self, model_dim, num_layers):
        if Mamba2Mixer is not None:
            return Mamba2Mixer(
                d_model=model_dim,
                n_layers=num_layers,
                d_state=self.d_state,
                d_conv=self.d_conv,
                headdim=self.headdim,
                expand=self.expand,
                dropout=self.dropout,
            )
        if Mamba1 is not None and Mamba1Config is not None:
            return Mamba1(
                Mamba1Config(
                    d_model=model_dim,
                    n_layers=num_layers,
                    d_state=self.d_state,
                    d_conv=self.d_conv,
                    expand_factor=self.expand,
                )
            )
        raise RuntimeError(
            "Failed to import any Mamba version - this should never happen",
        )



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import math
import os
from abc import abstractmethod
from typing import Any, Iterator, Protocol

import torch
from pydantic import Field
from torch.optim import Adam, Optimizer  # type: ignore
from torch.optim.lr_scheduler import CosineAnnealingLR, LinearLR, LRScheduler, SequentialLR

from mblm import MBLM, MBLMModelConfig, MBLMReturnType
from mblm.data.dataset.clevr import Clevr
from mblm.data.dataset.pg19 import PG19
from mblm.data.datasets import DistributedDataset
from mblm.data.types import BatchWithLossMask, ModelMode
from mblm.model.embeddings import MBLM_TOKEN_EMB_MIGRATION
from mblm.train.core.config import (
    CoreIoConfig,
    CoreModelParams,
    CoreTrainConfig,
    GenericEntryConfig,
    GenericOutputConfig,
)
from mblm.train.core.trainer import CoreTrainer
from mblm.utils.distributed import process_group
from mblm.utils.logging import create_logger, shutdown_log_handlers
from mblm.utils.misc import count_params


class TrainMBLMParams(MBLMModelConfig, CoreModelParams):
    """
    Combine the params required by the MBLM model and the trainer.
    """

    pass


class TrainMBLMIoConfig(CoreIoConfig):
    """
    Custom io settings on top of the core/required parameters
    """

    dataset_dir: str = Field(description="Path to the dataset folder")
    dataset_id: str = Field(description="The unique identifier of the dataset")
    dataset_args: dict[str, Any] | None = Field(
        default=None, description="Optional arguments passed to the dataset"
    )
    description: str | None = Field(default=None, description="A description for this experiment")


class TrainOutputConfig(GenericOutputConfig[TrainMBLMParams, CoreTrainConfig, TrainMBLMIoConfig]):
    """
    A convenience class that can be used directly to parse any output generated
    from training a MBLM.
    """

    pass


class TrainEntryConfig(GenericEntryConfig[TrainMBLMParams, CoreTrainConfig, TrainMBLMIoConfig]):
    """
    Class used to parse all required input configuration for training an MBLM.
    """

    pass


class MBLMTrainerDatasetImpl(Protocol):
    """
    Any dataset object implementing this protocol can be used with the MBLM trainer.
    """

    @staticmethod
    @abstractmethod
    def from_train_entry_config(
        config: TrainEntryConfig, mode: ModelMode, worker_id: int, num_workers: int
    ) -> DistributedDataset[BatchWithLossMask]:
        """
        How to parse a training config to a dataset.
        """
        ...

    @staticmethod
    @abstractmethod
    def supports_test_mode() -> bool:
        """
        Whether or not this dataset supports a test mode. Some datasets might not
        expose the answers in their test set so we cannot evaluate a model on it.
        Override if necessary
        """
        ...


class DatasetRegistry(dict[str, MBLMTrainerDatasetImpl]):
    """
    Dataset registry that allows (custom) datasets to be registered for usage
    with the MBLM trainer. Any dataset used with the MBLM trainer must be
    registered first.
    """

    def register(self, dataset_id: str, klass: MBLMTrainerDatasetImpl) -> None:
        """
        Register a dataset that implements the methods required by
        `MBLMTrainerDatasetImpl`.
        """
        return self.update({dataset_id: klass})

    def retrieve(self, dataset_id: str) -> MBLMTrainerDatasetImpl:
        """
        Retrieve a dataset. Will raise a `KeyError` if the dataset's id cannot
        be found.
        """
        return self[dataset_id]


class MegabyteTrainer(
    CoreTrainer[
        MBLM,
        BatchWithLossMask,
        TrainMBLMParams,
        CoreTrainConfig,
        CoreIoConfig,
    ]
):
    def init_model(self):
        return MBLM(
            MBLMModelConfig(
                # number of tokens
                num_tokens=self.config.params.num_tokens,
                # transformer model dimension (global, local)
                hidden_dims=tuple(self.config.params.hidden_dims),
                # sequence length (global, local)
                seq_lens=tuple(self.config.params.seq_lens),
                pad_token_id=self.config.params.pad_token_id,
                num_layers=tuple(self.config.params.num_layers),
                train_checkpoint_chunks=self.config.params.train_checkpoint_chunks,
                block=self.config.params.block,
            )
        )

    def model_forward(self, model, batch, device) -> torch.Tensor:
        inputs, loss_mask = batch
        inputs = inputs.to(device)
        loss_mask = loss_mask.to(device)
        loss: torch.Tensor = model.forward(
            inputs, return_type=MBLMReturnType.LOSS, loss_mask=loss_mask
        )
        return loss

    def configure_optimizer(self, parameters: Iterator[torch.nn.Parameter]) -> Optimizer:
        return Adam(
            parameters,
            lr=self.config.train.learning_rate,
            betas=(0.9, 0.95),
        )

    def configure_scheduler(self, optimizer, local_gradient_steps) -> LRScheduler:
        warmup_steps = math.floor(local_gradient_steps * self.config.train.warmup_steps_perc)
        linear = LinearLR(
            optimizer,
            total_iters=warmup_steps,
            start_factor=0.1,
            end_factor=1,
        )
        cosine_iters = local_gradient_steps - warmup_steps
        cosine = CosineAnnealingLR(optimizer, T_max=cosine_iters)
        return SequentialLR(
            optimizer,
            [linear, cosine],
            milestones=[warmup_steps],
        )

    def configure_run_id(self) -> str:
        return os.getenv("JOB_ID") or super().configure_run_id()

    def configure_count_parameters(self, model):
        return count_params(model)

    def migrate_embeddings_if_enabled(self):
        # older versions - the pg19 pretrained models - of mblm may have been trained
        # without modality tokens - provide this map to migrate the embeddings.
        # enabled via yaml config
        return MBLM_TOKEN_EMB_MIGRATION


dataset_registry = DatasetRegistry()

dataset_registry.register("pg19", PG19)
dataset_registry.register("clevr", Clevr)


def train_mblm(config: TrainEntryConfig) -> None:
    log = create_logger(__name__, log_dir=config.io.output_dir)

    try:
        with process_group(backend="nccl") as run_vars:
            dataset = dataset_registry.retrieve(config.io.dataset_id)

            train_dataset = dataset.from_train_entry_config(
                config,
                mode=ModelMode.TRAIN,
                worker_id=run_vars.local_rank,
                num_workers=run_vars.world_size,
            )
            valid_dataset = dataset.from_train_entry_config(
                config,
                mode=ModelMode.VALID,
                worker_id=0,
                num_workers=1,
            )

            trainer = MegabyteTrainer(config, run_vars=run_vars)
            best_model = trainer.train(train_dataset, valid_dataset)

            supports_test_mode = dataset.supports_test_mode()
            if best_model and supports_test_mode:
                test_dataset = dataset.from_train_entry_config(
                    config,
                    mode=ModelMode.TEST,
                    worker_id=0,
                    num_workers=1,
                )
                trainer.test(test_dataset, best_model)
    except Exception as error:
        log.fatal(error, exc_info=True)
        shutdown_log_handlers()



from __future__ import annotations

__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import json
import os
import random
from pathlib import Path
from typing import TYPE_CHECKING, Generator, Literal, TypedDict, overload

import torch
from pydantic import BaseModel, ConfigDict, Field, field_validator
from typing_extensions import Unpack

from mblm.data.datasets import DistributedDataset, DistributedDatasetConfig
from mblm.data.types import BatchWithLossMask, ModelMode
from mblm.data.utils import (
    Bytes,
    ColorSpace,
    ImagePipeline,
    Tokenizer,
    TokenizerOptions,
    shift_remap_tensor,
    target_loss_mask,
)

if TYPE_CHECKING:
    from mblm.train.mblm import TrainEntryConfig


class ClevrFunction(TypedDict):
    function: str
    inputs: list[int]
    value_inputs: list[str]


class ClevrQuestion(TypedDict):
    question: str
    answer: str
    image_filename: str
    question_family_index: int
    program: list[ClevrFunction]


class ClevrModelGeneration(BaseModel):
    id_model: str
    sample_idx: int
    question: str
    question_type: str
    answer_gen: list[int]
    answer_truth: list[int]
    ce: float
    timestamp: str


class ClevrOptionalArgs(BaseModel):
    model_config = ConfigDict(extra="forbid")

    target_mode: Literal["qiqa", "a"]
    qiqa_loss_mask: tuple[float, float, float, float]
    answer_categorical: bool = False
    resize_to_h: int | None = None
    resize_to_w: int | None = None
    crop_h_perc: float | None = None
    crop_w_perc: float | None = None
    downsample_channels: int | None = None
    shift_channels_start: int | None = None
    eom_token_id: int | None = None
    som_text_token_id: int | None = None
    som_image_token_id: int | None = None
    enable_jpeg_stream_with_quality: int | None = Field(
        default=None,
        description="When `None`, the image is represented as an array of pixel data. If an integer, the ordinary image processing is applied, however, the final image will consist of the raw, compressed JPEG byte stream with the specified quality.",
    )
    jpeg_in_rgb_color_space: bool = Field(
        default=True,
        description="If `True`, save the JPEG stream in RGB color space instead of YCbCr. This setting only has an effect if `enable_jpeg_stream_with_quality` is not `None`.",
    )

    @field_validator("enable_jpeg_stream_with_quality")
    @classmethod
    def assert_is_jpeg_quality(cls, v: int | None) -> int | None:
        """
        We compress the Clevr input PNG images via Pillow. This ensures the
        quality parameter for JPEG ranges from 0 to 95, where 0 is maximum
        compression and 95 least compression. See
        https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#jpeg-saving
        """
        if v is not None:
            assert 0 <= v <= 95
        return v


class Clevr(DistributedDataset[BatchWithLossMask]):
    """
    https://cs.stanford.edu/people/jcjohns/clevr/

    The `data_dir` is expected to be the of the exact structure as the
    original dataset, although the .txt files are not strictly needed:
        ├── COPYRIGHT.txt
        ├── LICENSE.txt
        ├── README.txt
        ├── images
        ├── questions
        └── scenes
    Note: Clevr does not include answers to test questions
    """

    ANSWER_CLASS_MAP: dict[str, str] = {
        "0": "0",
        "1": "1",
        "2": "2",
        "3": "3",
        "4": "4",
        "5": "5",
        "6": "6",
        "7": "7",
        "8": "8",
        "9": "9",
        "10": "A",
        "blue": "B",
        "brown": "C",
        "cube": "D",
        "cyan": "E",
        "cylinder": "F",
        "gray": "G",
        "green": "H",
        "large": "I",
        "metal": "J",
        "no": "K",
        "purple": "L",
        "red": "M",
        "rubber": "N",
        "small": "O",
        "sphere": "P",
        "yellow": "Q",
        "yes": "R",
    }

    QUESTION_TYPES: dict[str, str] = {
        "exist": "exists",
        "count": "count",
        # compare integer
        "equal_integer": "compare_integer",
        "less_than": "compare_integer",
        "greater_than": "compare_integer",
        # query attribute
        "query_color": "query_attribute",
        "query_material": "query_attribute",
        "query_size": "query_attribute",
        "query_shape": "query_attribute",
        # compare attribute
        "equal_size": "compare_attribute",
        "equal_material": "compare_attribute",
        "equal_shape": "compare_attribute",
        "equal_color": "compare_attribute",
    }

    # common (transposed) image shape for all modes
    IMAGE_SHAPE_C_W_H = 3, 480, 320
    MAX_QUESTION_LEN_BYTES = 205  # 203 for validation
    MAX_ANSWER_LEN_BYTESN = 8

    # number of question-answer pairs in the test/validation set
    NUM_ITEMS_TRAIN = 699_989
    NUM_ITEMS_VALID = 149_991

    # technically, Clevr images are RGBA - we drop the alpha channel
    image_color_space = ColorSpace.RGB

    def __init__(
        self,
        data_dir: str | Path,
        mode: ModelMode,
        pad_token_id: int,
        optional_args: ClevrOptionalArgs,
        **config: Unpack[DistributedDatasetConfig],
    ):
        root = Path(data_dir)
        self.mode = mode
        self.args = optional_args

        self.tokenize = Tokenizer(
            TokenizerOptions(
                pad_token_id=pad_token_id,
                eom_token_id=optional_args.eom_token_id,
                som_text_token_id=optional_args.som_text_token_id,
                som_image_token_id=optional_args.som_image_token_id,
            )
        ).pipeline

        if mode == ModelMode.TRAIN:
            questions_path = "questions/CLEVR_train_questions.json"
            img_root = "images/train"
        elif mode == ModelMode.VALID:
            questions_path = "questions/CLEVR_val_questions.json"
            img_root = "images/val"
        else:
            questions_path = "questions/CLEVR_test_questions.json"
            img_root = "images/test"
        with (root / questions_path).open() as file:
            self.entries: list[ClevrQuestion] = json.load(file)["questions"]
            self.images_root = root / img_root
        super().__init__(
            data_size=len(self.entries),
            is_sequential=False,
            **config,
        )

    @staticmethod
    def from_train_entry_config(
        config: TrainEntryConfig,
        mode: ModelMode,
        worker_id: int,
        num_workers: int,
    ) -> DistributedDataset[BatchWithLossMask]:
        # cannot pass None to model_validate
        optional_args = ClevrOptionalArgs.model_validate(config.io.dataset_args or dict())
        return Clevr(
            data_dir=config.io.dataset_dir,
            mode=mode,
            pad_token_id=config.params.pad_token_id,
            seq_len=config.params.input_seq_len,
            worker_id=worker_id,
            num_workers=num_workers,
            optional_args=optional_args,
        )

    @staticmethod
    def supports_test_mode():
        return False

    def process_and_flatten_img(self, image: ImagePipeline) -> torch.Tensor:
        """
        Process an image sample:

        1. Crop
        2. Resize
        3. Discretize/downsample channels
        4. If JPEG compression is enabled, return a tensor with data from the
           raw JPEG stream
        """
        if (h := self.args.crop_h_perc) and (w := self.args.crop_w_perc):
            image.crop((h, w))
        # resize
        if (w := self.args.resize_to_w) and (h := self.args.resize_to_h):
            image.resize((w, h))
        # discretize/downsample
        if disc_channels := self.args.downsample_channels:
            image.downsample_channels(disc_channels)
            # if we discretize, we optionally shift
            if shift_start := self.args.shift_channels_start:
                image_tensor, _, _ = shift_remap_tensor(image.to_tensor(), range_start=shift_start)
                image = ImagePipeline(image_tensor, self.image_color_space)
        if q := self.args.enable_jpeg_stream_with_quality:
            # already flattened
            keep_rgb = self.args.jpeg_in_rgb_color_space
            image_tensor = image.to_jpeg_buffer(q, keep_rgb).to_tensor()
        else:
            image_tensor = image.to_tensor().flatten()
        return image_tensor

    def get_sample_raw(self, from_idx: int) -> ClevrQuestion:
        """
        Get a raw sample as a (question, answer, image) tuple with no
        tokenization or preprocessing applied.
        """
        if self.mode == ModelMode.TEST:
            raise ValueError("Clevr dataset does not support testing!")
        return self.entries[from_idx]

    def iter_images(
        self, shuffle: bool = False, max_items: int | None = None
    ) -> Generator[torch.Tensor, None, None]:
        """
        For every image in Clevr, there are multiple questions and answers. Use
        this iterator if you want to iterate over images and yield a unique
        image each time. No preprocessing is applied.
        """
        image_lst = os.listdir(self.images_root)
        if shuffle:
            random.shuffle(image_lst)
        if max_items is not None:
            image_lst = image_lst[:max_items]
        for img_path in image_lst:
            yield ImagePipeline(self.images_root / img_path, self.image_color_space).to_tensor()

    @overload
    def iter(
        self, *, shuffle: bool = ..., max_items: int | None = ..., raw: Literal[True]
    ) -> Generator[tuple[int, ClevrQuestion], None, None]: ...
    @overload
    def iter(
        self, shuffle: bool = ..., max_items: int | None = ..., raw: Literal[False] = ...
    ) -> Generator[tuple[int, tuple[str, str, ImagePipeline]], None, None]: ...
    def iter(
        self, shuffle: bool = False, max_items: int | None = None, raw: bool = False
    ) -> Generator[tuple[int, tuple[str, str, ImagePipeline] | ClevrQuestion], None, None]:
        """
        Iterate over all the question, answer image tuples in Clevr (or raw
        entries if specified). While question/answer pairs unique, images may
        appear more than once. No preprocessing is applied.
        """
        entries_range = list(range(len(self.entries)))
        if shuffle:
            random.shuffle(entries_range)
        if max_items is not None:
            entries_range = entries_range[:max_items]
        for i in entries_range:
            s = self.get_sample_raw(i)
            if raw:
                yield i, s
            else:
                img = ImagePipeline(self.images_root / s["image_filename"], self.image_color_space)
                yield i, (s["question"], s["answer"], img)

    def get_sample_with_parts(
        self, from_idx: int
    ) -> tuple[
        torch.Tensor,
        torch.Tensor,
        tuple[torch.Tensor, torch.Tensor, torch.Tensor, int],
    ]:
        """
        Get a sample at an index with padding and tokenization applied. The
        returned tuple contains:

        1. The concatenated question, image, question, answer: (`q_i_q_a`)
        2. The loss mask with the same shape as `q_i_q_a`
        3. A tuple containing the original `q`, `i`, `a` parts that are
           concatenated in `q_i_q_a` and the size of the right padding applied
        """
        s = self.get_sample_raw(from_idx)
        question_str, answer_str = (s["question"], s["answer"])
        image_pipeline = ImagePipeline(
            self.images_root / s["image_filename"], self.image_color_space
        )
        # question and image as tensors, not yet tokenized
        question = Bytes.str_to_tensor(question_str)
        image = self.process_and_flatten_img(image_pipeline)

        # add special tokens - they will be converted from uint8 to long
        question = self.tokenize(question).with_som_text().with_eom().to_long_tensor()
        image = self.tokenize(image).with_som_image().with_eom().to_long_tensor()

        if self.args.answer_categorical:
            # override the n-length answer with a string of length 1
            answer_str = self.ANSWER_CLASS_MAP[answer_str]

        answer = Bytes.str_to_tensor(answer_str)
        answer = self.tokenize(answer).with_som_text().with_eom().to_long_tensor()

        # concat everything -> dtype long is preserved
        q_i_q_a = torch.concat([question, image, question, answer])
        len_original_q_i_q_a = len(q_i_q_a)

        # create a loss mask for q_i_q_a with the weights
        loss_mask = target_loss_mask(
            zip(
                [
                    question,
                    image,
                    question,
                    answer,
                ],
                self.args.qiqa_loss_mask,
            )
        )
        assert q_i_q_a.shape == loss_mask.shape
        # apply padding
        q_i_q_a = self.tokenize(q_i_q_a).pad_right_to(self.seq_len).to_long_tensor()
        loss_mask = self.tokenize(loss_mask).pad_right_to(self.seq_len).to_long_tensor()

        len_padding = len(q_i_q_a) - len_original_q_i_q_a

        return q_i_q_a, loss_mask, (question, image, answer, len_padding)

    def get_sample(self, from_idx: int) -> BatchWithLossMask:
        """
        Get a processed sample with a loss mask. This method is required by the
        DistributedDataset superclass.
        """
        q_i_q_a, loss_mask, _ = self.get_sample_with_parts(from_idx)
        return q_i_q_a, loss_mask



from __future__ import annotations

__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


import logging
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import Literal, TypeAlias

import numpy as np
import torch
from numpy.typing import NDArray
from PIL import Image
from PIL.Image import Image as PILImage

from mblm.data.utils import FileStream

# when the logging module is imported during runtime, PIL picks this up and
# pollutes sys.stdout because it uses a low log level. disable this. see
# https://github.com/camptocamp/pytest-odoo/issues/15
logging.getLogger("PIL").setLevel(logging.WARNING)


# https://pillow.readthedocs.io/en/stable/handbook/concepts.html#concept-modes
class ColorSpace(str, Enum):
    RGB = "RGB"
    GRAY = "L"


BinMode = Literal["lower", "upper", "mean"]
_BinModeValues: dict[str, tuple[float, float]] = {
    "lower": (0.0, 1.0),
    "upper": (1.0, 0.0),
    "mean": (0.5, 0.5),
}

ImagePipelineInput: TypeAlias = str | Path | PILImage | NDArray[np.uint8] | torch.Tensor


class ImagePipeline:
    """
    Create an image processing pipeline.

    - If the image is given as a Numpy array, the shape must be (H, W, C)
    - If given as a Torch tensor, the shape must be (C, H, W)
    - For grayscale images, the shape must be (H, W)

    As of now, this pipeline supports RGB and grayscale color spaces.
    """

    # we assume all images have a depth of 8 bit, which gives 2**8 color values
    MAX_NUM_CHANNELS = 256

    def __init__(self, image: ImagePipelineInput, color_space: ColorSpace):
        self._cs = color_space
        if isinstance(image, (str, Path)):
            image = Image.open(image)
        elif isinstance(image, (np.ndarray, torch.Tensor)):
            image = self._init_from_arraylike(image)

        self._image = image.convert(color_space.value)

    def _init_from_arraylike(self, image: NDArray[np.uint8] | torch.Tensor) -> PILImage:
        if isinstance(image, torch.Tensor):
            if self._cs == ColorSpace.RGB:
                image = image.permute(1, 2, 0)
            image = image.numpy()
        if (dtype := image.dtype) != np.uint8:
            raise ValueError(f"Can only process uint8 images, received {dtype} tensor")
        return Image.fromarray(image)

    def grayscale(self) -> ImagePipeline:
        if self._cs == ColorSpace.GRAY:
            return self
        self._image = self._image.convert(ColorSpace.GRAY.value)
        return self

    def resize(self, new_w_h: tuple[int, int]) -> ImagePipeline:
        """
        Resize an image to a new width and height (in pixels).
        """
        self._image = self._image.resize(new_w_h, resample=Image.Resampling.LANCZOS)
        return self

    def crop(self, factor_perc_h_w: tuple[float, float]) -> ImagePipeline:
        """
        Crop an image by (width, height) crop factors in percent.
        """
        width, height = self._image.size
        factor_h, factor_w = factor_perc_h_w

        margin_width = int(min(width * factor_w, width / 2))
        margin_height = int(min(height * factor_h, height / 2))

        left = margin_width
        top = margin_height
        right = width - margin_width
        bottom = height - margin_height

        self._image = self._image.crop((left, top, right, bottom))
        return self

    def downsample_channels(
        self, max_num_channel_values: int, bin_mode: BinMode = "mean"
    ) -> ImagePipeline:
        """
        Reduces the number of channel values in an image to a specified number.
        Channel values are effectively binned to `max_num_channel_values` bins.
        `min_bins_per_channel` must be greater than 0. If greater than 255,
        returns `self` unchanged.

        The strategy determines what value is assigned to the value in a bin, it
        can be `"upper"`, `"lower"` (where the values at the edges of bins are
        taken) or `"mean"` (which takes the average).

        By default, for `n` bins, values are mapped to the average bin width
        width. E.g., for `n=2` bins, all values between 0 and 127 are mapped to
        64. The values between 127 and 255 are mapped to 191.

        Note that due to floating point precision, channel values at the bin
        edges might fall into either the left or right bin.
        """
        # 255 bins is the maximum: 0 to 255
        if max_num_channel_values + 1 >= self.MAX_NUM_CHANNELS:
            return self

        if max_num_channel_values < 1:
            raise TypeError(
                f"Min number of bins must be positive, received {max_num_channel_values}"
            )

        # create bin decision edges from 0 to 255 at which to discretize values
        bin_borders = np.linspace(
            0, self.MAX_NUM_CHANNELS - 1, max_num_channel_values + 1, endpoint=True
        )
        assert len(bin_borders) == max_num_channel_values + 1
        # take any combination of two consecutive borders (n, n+1) to get the new values
        strategy = _BinModeValues[bin_mode]
        bin_values = np.convolve(bin_borders, strategy, mode="valid")
        # drop 255 - the last index is not needed for binning with np.digitize
        bin_borders = bin_borders[:-1]
        # discretize to get the indices to which midpoint values to map
        bin_indices = np.digitize(self.to_numpy(), bin_borders) - 1
        # extract discretized values at right positions
        binned_image = bin_values[bin_indices].astype(np.uint8)
        self._image = Image.fromarray(binned_image)
        return self

    def to_numpy(self) -> NDArray[np.uint8]:
        """
        Return the image as a numpy array with shape H, W, C.
        """
        return np.array(self._image, dtype=np.uint8)

    def to_tensor(self) -> torch.Tensor:
        """
        Return the image as a numpy array with shape C, H, W.
        """
        img_tensor = torch.from_numpy(self.to_numpy())
        if self._cs == ColorSpace.RGB:
            return img_tensor.permute(2, 0, 1)
        return img_tensor

    def to_image(self) -> PILImage:
        """
        Return the underlying Pillow image.
        """
        return self._image

    def to_jpeg_buffer(self, quality: int = 95, keep_rgb: bool = False) -> FileStream:
        """
        Compress the underlying image using JPEG compression and return a
        wrapped raw file buffer.

        Quality must be an integer between 0 (max compression) and 95 (minimal
        compression). See
        https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#jpeg-saving

        Note that you can transform the FileStream back into a Pillow image,
        however, you will need to manually convert to YCbCr color space.
        Although the choice of `keep_rgb` will be reflected in the buffer,
        reading it back into a Pillow image will automatically assume RGB color
        space.
        """
        if not 0 <= quality <= 95:
            raise ValueError(f"Invalid argument for quality: {quality}")
        buffer = BytesIO()
        self._image.save(buffer, "JPEG", quality=quality, keep_rgb=keep_rgb)
        return FileStream(buffer)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from io import BytesIO

import numpy as np
import torch
from numpy.typing import NDArray
from PIL import Image
from PIL.Image import Image as PILImage

DEFAULT_ENCODING: str = "utf8"


class Bytes:
    @staticmethod
    def tensor_to_str(t: torch.Tensor, encoding=DEFAULT_ENCODING) -> str:
        return bytes(t.tolist()).decode(encoding)

    @staticmethod
    def tensor_to_image(t: torch.Tensor) -> Image.Image:
        as_np = t.permute(1, 2, 0).numpy()
        return Image.fromarray(as_np)

    @staticmethod
    def str_to_tensor(s: str, encoding=DEFAULT_ENCODING) -> torch.Tensor:
        # frombuffer needs a mutable bytearray, not just bytes
        b = Bytes.str_to_bytearray(s, encoding)
        # note - doing
        # torch.frombuffer(bytearray(s.encode("utf8")), dtype=torch.uint8)
        # is more than 2x as fast as doing
        # torch.tensor(bytearray(s.encode("utf8")), dtype=torch.uint8)
        return torch.frombuffer(b, dtype=torch.uint8)

    @staticmethod
    def str_to_bytearray(s: str, encoding=DEFAULT_ENCODING) -> bytearray:
        # faster than bytearray(string, encoding)
        return bytearray(Bytes.str_to_bytes(s, encoding))

    @staticmethod
    def str_to_bytes(s: str, encoding=DEFAULT_ENCODING) -> bytes:
        return s.encode(encoding)

    @staticmethod
    def bytes_to_str(b: bytes | bytearray, encoding=DEFAULT_ENCODING) -> str:
        return b.decode(encoding)

    @staticmethod
    def byte_list_to_str(b: list[int], encoding=DEFAULT_ENCODING) -> str:
        return bytes(b).decode(encoding)

    @staticmethod
    def byte_list_to_str_safe(b: list[int], default: str, encoding=DEFAULT_ENCODING) -> str:
        try:
            return bytes(b).decode(encoding)
        except Exception:
            return default

    @staticmethod
    def bytes_to_tensor(b: bytes | bytearray) -> torch.Tensor:
        if isinstance(b, bytes):
            # cannot pass bytes directly to torch because bytes are not mutable.
            # wait for https://github.com/pytorch/pytorch/issues/69491
            return torch.tensor(memoryview(b), dtype=torch.uint8)
        return torch.frombuffer(b, dtype=torch.uint8)


class FileStream:
    """
    Wrapper around an ordinary file stream.

    For details on how the buffers are best converted, check
    https://stackoverflow.com/questions/61319551/when-should-one-use-bytesio-getvalue-instead-of-getbuffer
    """

    def __init__(self, buffer: BytesIO):
        self._buffer = buffer

    def to_numpy(self) -> NDArray[np.uint8]:
        """
        Return the buffer as a Numpy uint8 array.
        """
        return np.frombuffer(self._buffer.getvalue(), dtype=np.uint8)

    def to_tensor(self) -> torch.Tensor:
        """
        Return the buffer as a Torch uint8 Tensor.
        """
        return torch.frombuffer(self._buffer.getbuffer(), dtype=torch.uint8)

    def to_buffer(self) -> bytes:
        """
        Return the buffer as bytes.
        """
        return self._buffer.getvalue()

    def to_image(self) -> PILImage:
        """
        Return the buffer as a Pillow image.

        Because the underlying buffer may represent any image format with any
        color space, it is up to you to convert it accordingly.
        """
        return Image.open(self._buffer)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


from abc import ABC, abstractmethod
from typing import Any, Callable, Literal

import torch
from pydantic import BaseModel, Field


class StageBlock(ABC, BaseModel):
    """
    Configuration for a single model block at an MBLM stage.
    """

    block_type: str = Field(
        description="A name for the block for easy identification, which can be any string name"
    )

    pos_emb_type: Literal["fixed", "rope"] | None = Field(
        default=None,
        description="The type of positional embedding to add to tokens of the stage block",
    )

    @abstractmethod
    def to_model(
        self,
        model_dim: int,
        num_layers: int,
    ) -> torch.nn.Module: ...

    """
    An abstract method that creates a `torch.nn.Module` from the stage block
    configuration.
    """


class StageBlockRegistry(set[type[StageBlock]]):
    """
    Stage block registry that allows (custom) stage blocks to be registered for
    usage with MBLM. Blocks only need to be registered when they require parsing
    from a YAML config file. When a MBLM configuration is read from a YAML
    config file, the registry will try to find a matching stage block
    implementation.
    """

    def register(self, stage_block_klass: type[StageBlock]) -> None:
        """
        Register a stage block class so it can be validated from a YAML
        configuration file.
        """
        return self.add(stage_block_klass)

    def try_parse(
        self,
        data: Any,
        parse_func: Callable[[type[StageBlock], Any], StageBlock],
    ):
        """
        Try to parse configuration data to a registered stage block class.
        """
        for klass in self:
            try:
                return parse_func(klass, data)
            except Exception:
                pass

        raise ValueError(f"Coult not parse data to any of {self}")



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import logging
import math
from typing import Iterable, Literal, Sequence, cast, overload

import torch
import torch.nn.functional as F  # noqa: N812
from einops import pack, rearrange, repeat, unpack
from einops.layers.torch import Rearrange
from torch import nn
from torch.utils.checkpoint import checkpoint
from tqdm import tqdm

from mblm.model.block import StageBlock
from mblm.model.config import MBLMModelConfig, MBLMReturnType
from mblm.model.utils import ByteToUtf8Streamer, RoPE, gumbel_sample, top_k

"""
Wording:
    - A hierarchy consists of n stages
    - Stage 1 to n-1 corresponds to global models
    - Stage n corresponds to the local model
Inline comment abbreviations:
    -------------------------------------------------------------------
    Global notation
    -------------------------------------------------------------------
    - V        Vocabulary size (256 for bytes)
    - B        Batch size
    - L        The input sequence length
    - P_n      The sequence length/number of patches at any stage n
    - D_n      The model dimension at any stage n
"""


class MBLM(nn.Module):
    """
    Multiscale Byte Language Model is a hierarchical byte-level
    sequence-to-sequence model for multimodal tasks.

    Based on a refined fork of https://github.com/lucidrains/MEGABYTE-pytorch
    (version 0.3.5).
    """

    def __init__(self, cfg: MBLMModelConfig):
        super().__init__()

        assert len(cfg.hidden_dims) == len(cfg.num_layers) == len(cfg.seq_lens)
        if cfg.train_checkpoint_chunks:
            assert len(cfg.hidden_dims) - 1 == len(cfg.train_checkpoint_chunks)

        self.num_stages = len(cfg.hidden_dims)
        self.seq_lens = cfg.seq_lens
        self.pad_token_id = cfg.pad_token_id
        self.checkpoint_chunks: list[int | None] = (
            [None] + cfg.train_checkpoint_chunks
            if cfg.train_checkpoint_chunks
            else [None] * self.num_stages
        )

        self.start_tokens = nn.ParameterList(
            [nn.Parameter(torch.randn(model_dim)) for model_dim in cfg.hidden_dims]
        )

        self.pos_embs = self._init_positional_embeddings(
            cfg.hidden_dims, cfg.seq_lens, cfg.stage_blocks
        )

        self.token_embs_rev = self._init_token_embeddings(
            cfg.hidden_dims, cfg.seq_lens, cfg.num_tokens, cfg.pad_token_id
        )

        self.stage_models, self.to_next_stage_proj = self._init_models_at_stages(
            cfg.hidden_dims, cfg.seq_lens, cfg.num_layers, cfg.stage_blocks
        )

        self.to_logits = nn.Linear(cfg.hidden_dims[-1], cfg.num_tokens)

    @classmethod
    def _init_positional_embeddings(
        cls,
        model_dims: Iterable[int],
        seq_lens: Iterable[int],
        blocks: Iterable[StageBlock],
    ) -> nn.ModuleList:
        """
        Create positional embeddings for each patch.
        """
        modules = nn.ModuleList()
        for model_dim, seq_len, block in zip(model_dims, seq_lens, blocks):
            if not block.pos_emb_type:
                modules.append(nn.Identity())
            elif block.pos_emb_type == "fixed":
                modules.append(nn.Embedding(seq_len, model_dim))
            elif block.pos_emb_type == "rope":
                modules.append(
                    RoPE(
                        model_dim,
                        # only used for caching, will not fail if we're feeding
                        # in longer sequences - cache will be rebuilt with the
                        # new sequence length
                        max_seq_len=seq_len,
                    )
                )
        return modules

    @classmethod
    def _init_token_embeddings(
        cls,
        model_dims: Sequence[int],
        seq_lens: Sequence[int],
        vocab_size: int,
        pad_token_id: int,
    ) -> nn.ModuleList:
        """
        Embed the tokens for each stage (in reverse order).
        """
        local_dim = model_dims[-1]
        token_embs_rev = nn.ModuleList(
            [nn.Embedding(vocab_size, local_dim, padding_idx=pad_token_id)]
        )
        patch_size = 1
        for model_dim, seq_len in zip(
            # all except the local model
            reversed(model_dims[:-1]),  # (D_n-1, ..., D_1)
            reversed(seq_lens[1:]),  # (P_2, ..., P_n)
        ):
            # for the global models, fuse the embedding and patch projection
            # step
            patch_size *= seq_len
            token_embs_rev.append(
                nn.Sequential(
                    nn.Embedding(vocab_size, local_dim, padding_idx=pad_token_id),
                    Rearrange("... r d -> ... (r d)"),
                    nn.LayerNorm(patch_size * local_dim),
                    nn.Linear(patch_size * local_dim, model_dim),
                    nn.LayerNorm(model_dim),
                )
            )
        return token_embs_rev

    @classmethod
    def _init_models_at_stages(
        cls,
        model_dims: Sequence[int],
        seq_lens: Sequence[int],
        num_stage_layers: Iterable[int],
        blocks: Iterable[StageBlock],
    ) -> tuple[nn.ModuleList, nn.ModuleList]:
        """
        Initialize the models and next-stage projections for each stage.
        """
        stage_models = nn.ModuleList([])
        to_next_stage_proj = nn.ModuleList([])

        for block, model_dim, num_layers, next_model_dim, next_seq_len in zip(
            blocks,
            model_dims,
            num_stage_layers,
            tuple(model_dims[1:]) + (None,),
            tuple(seq_lens[1:]) + (None,),
        ):
            stage_models.append(block.to_model(model_dim=model_dim, num_layers=num_layers))

            proj: torch.nn.Module = nn.Identity()

            if next_model_dim and next_seq_len:
                proj = nn.Sequential(
                    Rearrange("b ... d -> b (...) d"),
                    nn.Linear(model_dim, next_model_dim * next_seq_len),
                    Rearrange("b m (n d) -> (b m) n d", n=next_seq_len),
                )

            to_next_stage_proj.append(proj)
        return stage_models, to_next_stage_proj

    def forward_empty(self, batch_size: int) -> torch.Tensor:
        """
        Take care of special case where you sample from input of length 0 (start
        token only).
        """

        prev_stage_tokens_repr: torch.Tensor | None = None
        tokens: torch.Tensor = torch.empty(0)

        for stage_start_tokens, stage_model, proj in zip(
            self.start_tokens, self.stage_models, self.to_next_stage_proj
        ):
            tokens = repeat(stage_start_tokens, "d -> b 1 d", b=batch_size)

            if prev_stage_tokens_repr is not None:
                tokens = tokens + prev_stage_tokens_repr[..., : tokens.shape[-2], :]

            tokens = stage_model(tokens)
            prev_stage_tokens_repr = proj(tokens)

        return self.to_logits(tokens)

    @overload
    def forward(
        self,
        input_ids: torch.Tensor,
        *,
        return_type: Literal[MBLMReturnType.LOSS_LOGITS] = ...,
        loss_mask: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor]: ...
    @overload
    def forward(
        self,
        input_ids: torch.Tensor,
        *,
        return_type: Literal[MBLMReturnType.LOSS, MBLMReturnType.LOGITS] = ...,
        loss_mask: torch.Tensor | None = None,
    ) -> torch.Tensor: ...

    def forward(
        self,
        input_ids: torch.Tensor,
        *,
        return_type: MBLMReturnType = MBLMReturnType.LOSS,
        loss_mask: torch.Tensor | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
        """
        A single forward pass.

        Args:
            input_ids: The token ids as torch.LongTensor in shape (B, L)
            return_type: What to return - the loss, the logits or both.
            loss_mask: An optional masking tensor that enables interpolation between
                self-supervised and supervised learning. It determines which tokens in the
                prediction should contribute to the loss with what weight and should have
                the same shape as `input_ids`. For example, if `loss_mask` is a tensor
                `torch.ones(input_ids)`, we learn to predict `input_ids` (self-supervised).
                In other scenarios, `input_ids` might be a flat tensor with `...question,
                ...answer`. In this case, one can provide a `loss_mask` in which the
                question tokens have a mask of 0 and the answer a mask of 1. The loss
                is then only computed on the answer, resulting in a supervised setting.
                By default, not providing a `loss_mask` is equivalent to a `loss_mask`
                consisting of all 1
        """
        batch_size = input_ids.shape[0]

        assert input_ids.ndim in {2, self.num_stages + 1}

        if input_ids.numel() == 0:
            return self.forward_empty(batch_size)

        if loss_mask is not None:
            assert loss_mask.shape == input_ids.shape

        flattened_dims = input_ids.ndim == 2
        flat_seq_len = input_ids.shape[-1]

        # if the input is given as (B, L), reshape and distribute it among the
        # hierarchy sequence lengths, filling up the inner dimensions
        # first. padding is applied so that the output shape is (B, P_1', P_2,
        # ..., P_n). for the largest possible L == prod(seq_lens), P_1' = P_1.
        # in all other cases, P_1' < P_1, meaning no padding is applied to the
        # global sequence P_1.
        #
        # here's two examples for a model model (P_1, P_2, P_3) = (5, 4, 3):
        #
        # input: (B, L) = (1, 13), output: (B, P_1', P_2, P_3) = (1, 2, 4, 3)
        # input: (B, L) = (1, 60), output: (B, P_1,  P_2, P_3) = (1, 5, 4, 3)
        #
        # in the 2nd example, L = prod(seq_lens) = 5 * 4 * 3 = 60 = P_1 = P_1'
        if flattened_dims:
            # pad/fill up all inner sequence lengths (all except most global)
            local_seq_lens = self.seq_lens[1:]
            multiple_of = math.prod(local_seq_lens)
            # use the complement of modulo - the difference to the next multiple
            # of multiple_of - to infer the right padding length
            padding = -flat_seq_len % multiple_of
            input_ids = F.pad(input_ids, (0, padding), value=self.pad_token_id)
            # reshape and infer the P_1' dimension
            input_ids = input_ids.reshape(batch_size, -1, *local_seq_lens)

        # make sure the above condition holds, i.e., P_1' <= P_1
        _P_1_prime, _P_1 = input_ids.shape[1], self.seq_lens[0]  # noqa: N806
        fixed_global_patch_encoding = isinstance(self.pos_embs[0], nn.Embedding)
        if fixed_global_patch_encoding:
            assert _P_1_prime <= _P_1, (
                f"Because you are using a fixed global patch embedding, "
                f"the input sequence length ({_P_1_prime}) "
                f"must be less than the first tuple element of seq_lens ({_P_1})"
            )

        token_embs_at_stages = [torch.empty(0) for _ in range(self.num_stages)]
        # at this stage, we're working with nested ids - hence, embed the bytes
        # for each stage in reverse order, starting from the local and ending at
        # the most global model. at each stage, add positional embeddings and
        # rerrange the input shape to match the dimension of the previous
        # (local) stage. with three stages:
        #
        # [0]: (B, P_1', D_1)
        # [1]: (B, P_1', P_2, D_2)
        # [2]: (B, P_1', P_2, P_3, D_3)
        for stage_idx, pos_emb, token_emb in zip(
            range(self.num_stages - 1, -1, -1),
            reversed(self.pos_embs),
            self.token_embs_rev,
        ):
            stage_token_embs: torch.Tensor = token_emb(input_ids)
            stage_seq_len = stage_token_embs.shape[-2]

            if isinstance(pos_emb, nn.Embedding):
                positions: torch.Tensor = pos_emb(
                    torch.arange(stage_seq_len, device=input_ids.device),
                )
                stage_token_embs = stage_token_embs + positions
            elif isinstance(pos_emb, RoPE):
                batch, *seq_lens, hidden_dim = stage_token_embs.shape
                # RoPE expects as input [batch, seq_len, num_heads, head_dim] ->
                # pack to artificial batch dim and add an empty head dimension
                stage_token_embs = stage_token_embs.view(-1, seq_lens[-1], 1, hidden_dim)
                stage_token_embs = pos_emb.forward(stage_token_embs)
                # reshape back to input
                stage_token_embs = stage_token_embs.view(batch, *seq_lens, hidden_dim)

            token_embs_at_stages[stage_idx] = stage_token_embs

            # skip rearranging for the most local model
            if stage_idx == self.num_stages - 1:
                continue
            input_ids = rearrange(input_ids, "... m n -> ... (m n)")

        # initials
        prev_stage_tokens_repr: torch.Tensor | None = None
        attended: torch.Tensor = torch.empty(0)

        for stage_idx in range(self.num_stages):
            stage_start_token = self.start_tokens[stage_idx]
            stage_seq_len = self.seq_lens[stage_idx]
            stage_emb_tokens = token_embs_at_stages[stage_idx]
            model = self.stage_models[stage_idx]
            checkpoint_chunk = self.checkpoint_chunks[stage_idx]
            proj = self.to_next_stage_proj[stage_idx]
            # for each stage n, going from global to local, pack the tokens into
            # a new tensor so that the last two dimensions correspond to the
            # sequence length and the dimension of that stage - later, the shape
            # is restored by the unpack operations.

            # the first dimension ("*"), after packing, corresponds to the
            # number of patches in the current stage (which can be considered a
            # batch for the stage because they are processed in parallel). the
            # patch size of model n-1 becomes the batch size of model n, and so
            # on
            stage_tokens, ps = pack([stage_emb_tokens], "* s d")
            patch_batch = stage_tokens.shape[0]  # "*" in the pack operation above

            # reshape the start tokens and prepend them to the stage tokens
            stage_start_token = repeat(stage_start_token, "d -> b 1 d", b=patch_batch)
            stage_tokens = torch.cat((stage_start_token, stage_tokens), dim=-2)
            # sum the previous hierarchy's representation
            if prev_stage_tokens_repr is not None:
                prev_stage_tokens_repr = F.pad(
                    # pad the second last dimensions (stage sequence) by 1 to
                    # the top to skip adding the embedding to the start token
                    # (i.e., addition with 0)
                    prev_stage_tokens_repr,
                    (0, 0, 1, 0),
                    value=0,
                )
                _start_token = stage_tokens[..., 0, :]
                stage_tokens = stage_tokens + prev_stage_tokens_repr
                assert _start_token.equal(stage_tokens[..., 0, :])
            # stage_tokens is now [B, P_n, D_n]. for the first stage, B is the
            # actual batch size whereas for the other stages, the batch size
            # corresponds to the sequence length of the previous hierarchy stage

            # skip checkpointing for the first global stage
            if checkpoint_chunk is None:
                attended = model.forward(stage_tokens)

            else:
                chunks: list[torch.Tensor] = []
                batch_chunks = torch.chunk(stage_tokens, checkpoint_chunk, dim=0)

                for batch_chunk in batch_chunks:
                    # although not explicitly needed, do not checkpoint for
                    # inference - this saves a little bit of time (around 0.1s /
                    # sequence for a sequence over 1M elements)
                    out = (
                        cast(
                            torch.Tensor,
                            checkpoint(
                                model.forward,
                                batch_chunk,
                                use_reentrant=False,
                            ),
                        )
                        if self.training
                        else model.forward(batch_chunk)
                    )

                    chunks.append(out)
                attended = torch.cat(chunks, dim=0)

            # the output and input share are equal -> by unpacking again, we
            # restore the initial shape
            attended = unpack(attended, ps, "* s d")[0]
            # project for next stage in the hierarchy, dropping the last patch:
            # from (..., P_n, D_n) to (..., P_n, P_n+1, D_n+1)
            prev_stage_tokens_repr = proj(attended[..., :-1, :])

        logits = self.to_logits.forward(attended)  # (B, P_1', P_2, ..., 1 + P_n, V)

        logits_out = logits

        if flattened_dims:
            # drop the start tokens and combine inner dimensions into one
            logits_out = rearrange(logits_out[..., 1:, :], "b ... v -> b (...) v")
            # remove the padding
            logits_out = logits_out[:, :flat_seq_len]

        if return_type == MBLMReturnType.LOGITS:
            return logits_out

        # the most local sequences still contain the preprended start tokens
        # (i.e., the logits) - extract this token once and later prepend it to
        # the flattened sequence
        first_start_token_logits_idx = (slice(None), *((0,) * (logits.ndim - 2)), slice(None))
        # extract it
        start_token_logits = logits[first_start_token_logits_idx]  # type: ignore

        # drop ALL start tokens and combine inner dimensions into one...
        logits_rearranged = rearrange(logits[..., 1:, :], "b ... v -> b (...) v")

        # spread across batches
        start_token_logits = rearrange(start_token_logits, "b v -> b 1 v")
        # right-shift by one by appending the start token to the flat sequence
        logits_rearranged = torch.cat((start_token_logits, logits_rearranged), dim=-2)
        # drop the last item to match lengths
        logits_rearranged = logits_rearranged[..., :-1, :]

        # rearrange for loss calculation: the k-dimensional loss expects (B, V,
        # L)
        preds = rearrange(logits_rearranged, "b l v -> b v l")
        targets = rearrange(input_ids, "b ... -> b (...)")

        # same shape as targets with 0 everywhere where the token equals the pad
        # token. this assumes the same pad token is used for intra-batch padding
        # (ensured by the datasets/dataloaders) as well as patch-padding
        # (ensured by bootstrapping MBLM with the right pad token id)
        loss_tensor: torch.Tensor = F.cross_entropy(
            preds,  # (B, V, L)
            targets,  # (B, L)
            ignore_index=self.pad_token_id,
            reduction="none",
        )

        # remove the hierarchy padding (ignored in the loss calculation) and
        # bring to same shape as input_ids
        loss_tensor = loss_tensor[:, :flat_seq_len]

        if loss_mask is not None:
            # potentially apply the loss mask. this does not involve
            # broadcasting as after slicing above, the mask and the loss tensor
            # have the exact same shape again
            loss_tensor *= loss_mask

        # after applying the mask, some elements might be 0 - they should not be
        # accounted for in the loss calculation
        nonzero_idxs = torch.nonzero(loss_tensor, as_tuple=True)
        loss = loss_tensor[nonzero_idxs].mean()
        if torch.isnan(loss):
            # special case when the loss is zero across target elements - should
            # theoretically never happen
            logging.fatal("Edge case detected, loss is nan")
            loss = torch.zeros_like(loss, requires_grad=True)

        if return_type == MBLMReturnType.LOSS:
            return loss
        return loss, logits_out

    @torch.inference_mode()
    def generate(
        self,
        prime: torch.Tensor | None = None,
        stream: ByteToUtf8Streamer | None = None,
        num_tokens_to_generate: int | None = None,
        end_of_gen_token_id: int = -1,
        temperature: float = 1.0,
        filter_thres: float = 1.0,
        enable_progress: bool = True,
    ) -> torch.Tensor:
        """
        Autoregressively generate tokens.

        Args:
            prime: The prompt as a 1D tensor of type torch.long. Batches are
                not supported for now
            stream: Instead of generating and then returning the sequene, stream
                the intermediate results. Only possible when the bytes represent
                UTF-8 encoded text
            num_tokens_to_generate: If set, ignore the context window and
                generate an exact amount of tokens. If `None`, generate up to
                the maximum possible sequence length.
            temperature: A float in the range (0, 1] that determines the
                temperature. 1 means deterministic, 0.1 very random. Has
                no effect if `filter_thres` is set to 1
            filter_thres: A float in the range [0, 1] to determine the
                threshold for the top k logits. 1 means the token with
                the maximum logit is sampled.
        """
        if prime is not None:
            assert prime.ndim == 1, "Batch mode not supported"
            prime = prime.unsqueeze(0)
        else:
            device = next(self.parameters()).device
            prime = torch.empty((1, 0), dtype=torch.long, device=device)

        total_iters = num_tokens_to_generate or math.prod(self.seq_lens) - prime.shape[-1]

        iter_rng = range(total_iters)
        iterator = tqdm(iter_rng, leave=False) if (stream is None and enable_progress) else iter_rng

        sequence = prime
        for _ in iterator:
            logits = self.forward(sequence, return_type=MBLMReturnType.LOGITS)[:, -1]
            logits = top_k(logits, thres=filter_thres)
            sampled = gumbel_sample(logits, dim=-1, temperature=temperature)
            sequence = torch.cat((sequence, rearrange(sampled, "b -> b 1")), dim=-1)

            newest_token = int(sampled.item())

            if stream is not None:
                stream.write(newest_token)

            if newest_token == end_of_gen_token_id:
                break
        return sequence.squeeze()



from __future__ import annotations

__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


from dataclasses import dataclass

import torch
import torch.nn.functional as F  # noqa: N812


@dataclass
class TokenizerOptions:
    """
    Options for to be used for the Tokenizer class.

    Args:
        pad_token_id: ID of the padding token
        eom_token_id: End-of-modality token ID
            (can be considered a start-of-sentence-token in language modeling)
        som_text_token_id: Start-of-text-modality token ID
        som_image_token_id: Start-of-image-modality token ID
    """

    pad_token_id: int
    eom_token_id: int | None
    som_text_token_id: int | None
    som_image_token_id: int | None
    dtype = torch.long


@dataclass
class Tokenizer:
    """
    A preprocessor to add special tokens to 1D tensors.


    **Example**::

        pipeline = Tokenizer(
            TokenizerOptions(
                pad_token_id=10,
                eom_token_id=11,
                som_image_token_id=12,
                som_text_token_id=13,
            )
        ).pipeline
        input_t = torch.arange(0, 10)  # length 10
        input_t = pipeline(input_t).with_eom().with_som_text().pad_right_to(15).ok()

        assert input_t.size(0) == 15
        assert input_t[0].item() == 13  # som token
        assert input_t[11].item() == 11  # eom token
        assert input_t[12:].equal(torch.tensor([10, 10, 10]))  # padding token
    """

    options: TokenizerOptions

    def pipeline(self, tensor: torch.Tensor) -> TokenizerPipeline:
        """
        Start a piepline to process the 1D input tensor.
        """
        return TokenizerPipeline(tensor=tensor, options=self.options)


@dataclass
class TokenizerPipeline:
    tensor: torch.Tensor
    options: TokenizerOptions

    def __post_init__(self):
        if (dim := self.tensor.ndim) > 1:
            raise ValueError(f"Can only process 1D tensors, input is {dim}D")

    def _tensor_from_val(self, fill_value: int | None) -> torch.Tensor:
        if fill_value is None:
            return torch.empty((0), dtype=self.options.dtype)
        else:
            return torch.full((1,), fill_value=fill_value, dtype=self.options.dtype)

    def with_eom(self) -> TokenizerPipeline:
        filled = self._tensor_from_val(self.options.eom_token_id)
        self.tensor = torch.cat([self.tensor, filled])
        return self

    def with_som_text(self) -> TokenizerPipeline:
        filled = self._tensor_from_val(self.options.som_text_token_id)
        self.tensor = torch.cat([filled, self.tensor])
        return self

    def with_som_image(self) -> TokenizerPipeline:
        filled = self._tensor_from_val(self.options.som_image_token_id)
        self.tensor = torch.cat([filled, self.tensor])
        return self

    def pad_right_to(self, to_length: int, strict: bool = True) -> TokenizerPipeline:
        current_lengt = self.tensor.size(0)
        pad_right = to_length - current_lengt
        if strict and pad_right < 0:
            raise ValueError(
                f"Tensor at dim 0 (length {current_lengt}) larger than "
                f"desired padded size {to_length}"
            )
        self.tensor = F.pad(self.tensor.long(), (0, pad_right), value=self.options.pad_token_id)
        return self

    def to_long_tensor(self) -> torch.Tensor:
        return self.tensor.long()



from __future__ import annotations

__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""


from abc import ABC, abstractmethod
from typing import TypeVar

from torch.utils.data import Dataset
from typing_extensions import TypedDict, Unpack

_T = TypeVar("_T")


class DistributedDatasetConfig(TypedDict):
    seq_len: int
    worker_id: int
    num_workers: int


# TODO: Python 3.12, Self
class DistributedDataset(ABC, Dataset[_T]):
    """
    Custom Dataset class that distributes the data across a predefined number of
    workers. It implements all the methods needed to qualify as Dataset. See
    https://pytorch.org/docs/stable/data.html#map-style-datasets

    If `is_sequential` is true, then the underlying data is assumed as
    sequential chunks - a sequence - from an interable (e.g., a list of
    integers). If false, it is assumed that each sampling retrieves only one
    element.

    The most important parameter is `enable_offset`, which affects sampling of
    the underlying data: When set to `True`, the methods `self.offset_to` and
    `self.offset_one` enable dynamic shifting of the starting index when
    retrieving data sequences

    - `self.offset_to(offset)` sets the current sample offset to a specified
      value, allowing data to be accessed from different starting points within
      each sequence, introducing variability in the sampled data
    - `self.offset_one()` increments the sample offset by 1

    When `enable_offset` is false, the dataset fetches sequences starting from
    fixed positions based on the given index, with no offset applied (like a
    regular Dataset). Enabling the offset can be useful in tasks like sequence
    modeling where shifting the sequence start positions can improve the model's
    robustness or generalization by providing varied training inputs. See the
    test for this class for details on the effects
    """

    def __init__(
        self,
        data_size: int,
        is_sequential: bool,
        **kwargs: Unpack[DistributedDatasetConfig],
    ):
        num_workers, worker_id = kwargs["num_workers"], kwargs["worker_id"]
        seq_len = kwargs["seq_len"]

        assert num_workers > 0, "num_workers must be positive"
        assert (
            worker_id < num_workers
        ), f"worker_id ({worker_id}) must be smaller than num_workers ({num_workers})"

        worker_numel = data_size // num_workers
        if is_sequential:
            assert worker_numel >= seq_len, "Worker's data is too small"
        else:
            assert worker_numel >= 1

        self._worker_range_start = worker_id * worker_numel
        self._worker_range_end = self._worker_range_start + worker_numel

        self._worker_numel = worker_numel

        # public properties that may be accessed
        self.is_sequential = is_sequential
        self.seq_len = seq_len
        self.sample_offset = 0

    def offset_to(self, offset: int) -> DistributedDataset[_T]:
        """
        Offset for sequential sampling - has no effect if this `is_sequential`
        is false!
        """
        if not self.is_sequential:
            return self
        self.sample_offset = (offset) % self.seq_len
        return self

    def offset_one(self) -> DistributedDataset[_T]:
        return self.offset_to(self.sample_offset + 1)

    @abstractmethod
    def get_sample(self, from_idx: int) -> _T:
        """
        Abstract method that must be implemented to tell this superclass how to
        get a single data sample. By construction, a dataset is abstract and may
        hold different kinds of data such as a single Tensor containing
        sequences for autoregressive modelling or a tuple of (features, labels)
        Tensors.

        When you want to sample distinct items from your dataset, you can
        retrieve them directly at the index and use `self.seq_len` to pad or
        prepare your data. This is only enabled when the base class has been
        created with `is_sequential` set to false. If created with true, then
        you should sample from as `from_idx : from_idx + self.seq_len`
        """
        ...

    def __len__(self) -> int:
        """
        The total number of indexable elements in the dataset, e.g., the number
        of sequences. This property accounts for both sequence length and
        offsets and is thus strictly smaller or equal to `num_items`. It is
        always strictly positive. This number is used for the `__len__`
        implementation.
        """
        if self.is_sequential:
            available = self._worker_numel - self.sample_offset
            return max(available // self.seq_len, 0)
        else:
            return self._worker_numel

    def __getitem__(self, idx: int) -> _T:
        idx_mult = self.seq_len if self.is_sequential else 1
        from_idx = (idx * idx_mult) + self.sample_offset + self._worker_range_start
        return self.get_sample(from_idx)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from enum import Enum, auto
from itertools import repeat
from typing import Any, Sequence

from pydantic import BaseModel, computed_field, field_validator, model_validator

from mblm.model.block import StageBlock, StageBlockRegistry
from mblm.model.mamba import MambaBlockConfig
from mblm.model.transformer import TransformerBlockConfig

block_registry = StageBlockRegistry()
block_registry.register(TransformerBlockConfig)
block_registry.register(MambaBlockConfig)


class MBLMReturnType(str, Enum):
    LOGITS = auto()
    LOSS = auto()
    LOSS_LOGITS = auto()


class MBLMModelConfig(BaseModel):
    """
    General config for creating a MBLM model. For all iterables,
    the order corresponds to global to local stage from left to right.

    Params:
        num_tokens: The vocabulary size
        pad_token_id: Id of the padding token
        hidden_dims: The model's hidden dimensions at each stage
        num_layers: The number of layers at each stage
        seq_lens: The sequence length at each stage
        block: A Transformer or Mamba block
    """

    num_tokens: int
    pad_token_id: int
    hidden_dims: Sequence[int]
    num_layers: Sequence[int]
    seq_lens: Sequence[int]
    train_checkpoint_chunks: list[int] | None

    block: StageBlock | list[StageBlock]

    @field_validator("block", mode="before")
    @classmethod
    def parse_block(cls, value: Any) -> Any:
        """
        Try to parse the block to any registered stage block
        """
        given_as_list = isinstance(value, list)
        value = value if given_as_list else [value]
        parsed = []
        for block_data in value:
            if isinstance(block_data, StageBlock):
                parsed.append(block_data)
                continue
            parsed_block = block_registry.try_parse(
                block_data, lambda model, data: model.model_validate(data)
            )
            parsed.append(parsed_block)
        return parsed if given_as_list else parsed[0]

    @model_validator(mode="after")
    def validate_stages(self):
        is_equal_num_stages = len(self.hidden_dims) == len(self.num_layers) == len(self.seq_lens)
        if not is_equal_num_stages:
            raise ValueError("hidden_dims, num_layers and seq_lens are not of the same length")

        if isinstance(self.block, list) and len(self.block) != len(self.hidden_dims):
            raise ValueError(
                "When given as a list of blocks, must have the appropriate number of stages"
            )
        return self

    # https://docs.pydantic.dev/2.0/usage/computed_fields/
    @computed_field  # type: ignore[misc]
    @property
    def stage_blocks(self) -> list[StageBlock]:
        if isinstance(self.block, list):
            return self.block
        return list(repeat(self.block, len(self.hidden_dims)))



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import csv
import json
import logging
from pathlib import Path
from typing import Any, Callable, Generic, Iterable, NamedTuple, TypeAlias, TypeVar

import torch
import yaml
from filelock import FileLock
from pydantic import BaseModel
from torch import load as torch_load
from torch import save as torch_save
from torch.nn import Module as TorchModule
from torch.serialization import MAP_LOCATION


def _to_path(path: str | Path) -> Path:
    return Path(path) if isinstance(path, str) else path


# Although type parameter lists exist, they're only available in Python >=3.12.
# Use TypeVar and ParamSpec objects instead, which are backwards compatible with
# 3.10. See https://docs.python.org/3/library/typing.html#typing.ParamSpec
_T = TypeVar("_T")
_TBaseModel = TypeVar("_TBaseModel", bound=BaseModel)


def load_yml(
    path: str | Path, parse_to: type[_TBaseModel], try_yaml_suffixes: bool = False
) -> _TBaseModel:
    """
    Load any configuration from a (nested) yaml file.
    Args:
        path (str | Path): The path to the yaml file
        parse_to: An instance of a Pydantic model
        check_yaml_suffixes (bool=False): If `True`, try
            loading the file with either a `yaml` or
            `yml` extension
    Returns:
        A populated instance of the Pydantic model
    """
    path = _to_path(path)
    yaml_suffixes = {".yaml", ".yml"}
    assert path.suffix in yaml_suffixes, f"{path} is not a yaml file"
    try:
        with Path.open(path, "r") as file:
            yaml_data = yaml.safe_load(file)
            return parse_to.model_validate(yaml_data)
    except FileNotFoundError as exception:
        if try_yaml_suffixes:
            yaml_suffixes.remove(path.suffix)
            new_suffix = next(iter(yaml_suffixes))
            return load_yml(path.with_suffix(new_suffix), parse_to, try_yaml_suffixes=False)
        raise exception


def dump_yml(path: str | Path, data: BaseModel) -> Path:
    path = _to_path(path).with_suffix(".yaml")
    with Path.open(path, "w") as file:
        yaml.safe_dump(data.model_dump(), file)
    return path


_TNamedTuple = TypeVar("_TNamedTuple", bound=NamedTuple)

# dont pollute with package internal logs
logging.getLogger("filelock").setLevel(logging.WARNING)


class CSVWriter(Generic[_TNamedTuple]):
    def __init__(
        self,
        output_dir: str | Path,
        file_name: str,
        noop: bool = False,
    ) -> None:
        self.noop = noop
        if noop:
            return
        output_dir = _to_path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        self._file = (output_dir / file_name).with_suffix(".csv")

        with self._file.open("w", encoding="utf-8"):
            # clear existing file contents
            pass
        self._lock = FileLock(str(self._file) + ".lock")

    def _file_is_empty(self) -> bool:
        """Check if the file is empty."""
        return self._file.stat().st_size == 0

    def write_row(self, row: _TNamedTuple):
        """
        Write a row.

        Args:
            row: Named tuple to write.
        """
        if self.noop:
            return

        with self._lock:
            file_empty = self._file_is_empty()
            with self._file.open("a", encoding="utf-8") as f:
                writer = csv.writer(f)
                if file_empty:
                    # automatically write a header once
                    writer.writerow(row._fields)
                writer.writerow(row)


class NDJSONWriter(Generic[_TBaseModel]):
    def __init__(self, path: Path | str):
        self.path = _to_path(path)
        self.stream_positions: list[int] = []

    def write_line(self, obj: _TBaseModel) -> None:
        with self.path.open("a", encoding="utf8") as f:
            self.stream_positions.append(f.tell())
            f.write(obj.model_dump_json() + "\n")

    def remove_last_line(self) -> None:
        if len(self.stream_positions) == 0:
            return None
        last_position = self.stream_positions.pop()
        with self.path.open("a", encoding="utf8") as f:
            f.truncate(last_position)


def read_jsonl(
    path: Path | str,
    parse_lines_to: Callable[..., _T],
) -> list[_T]:
    with _to_path(path).open("r", encoding="utf8") as f:
        return [parse_lines_to(**json.loads(line)) for line in f.readlines()]


StateDict: TypeAlias = dict[str, Any]


def save_model_state(
    dir: str | Path,
    checkpoint_name: str,
    model: TorchModule | StateDict,
    loss: float,
) -> tuple[bool, Path]:
    """
    Save a model's state_dict to a checkpoint. Will overwrite a file with the same name.

    Args:
        dir (str | Path): If it does not exist, it will be created recursively
        checkpoint_name (str): The suffix will be set to `.pth` automatically.
        model (torch.nn.Module | StateDict): Any class that inherits from
            `torch.nn.Module` or the actual state dict obtained from `model.state_dict()`

    Returns:
        did_overwrite (bool): Whether or not the file was overwritten
    """
    path = _to_path(dir)
    path.mkdir(parents=True, exist_ok=True)
    file = path.joinpath(checkpoint_name).with_suffix(".pth")

    if isinstance(model, TorchModule):
        model = model.state_dict()
    did_overwrite = file.exists()
    snapshot = {"MODEL": model, "LOSS": loss}
    torch_save(snapshot, file)
    return did_overwrite, file


_TModule = TypeVar("_TModule", bound=TorchModule)


def _embedding_map_extend(
    src_state: dict[str, torch.Tensor],
    tgt_state: dict[str, torch.Tensor],
    map_extend_embeddings: set[str],
    notify: Callable[[str], Any],
) -> dict[str, torch.Tensor]:
    """
    Special loader for when we have module A with an embedding or linear layer
    of size (N, D) and module B with embedding or linear layer of size (N+i, D).
    This function transfers the source weights and biases into the corresponding
    positions in the target, leaving additional entries uninitialized or
    randomly initialized. Hence, it is required that the source size in A is
    smaller than the target size in B. At the moment, we only support this
    extending strategy, not shrinking.
    """
    # TODO: Allow init of newly added embeddings with mean, similar to
    # https://github.com/huggingface/transformers/blob/ccbd57a8b665fbb5b1d566c0b800dc6ede509e8e/src/transformers/modeling_utils.py#L2465
    for (src_key, src_tensor), (tgt_key, tgt_tensor) in zip(src_state.items(), tgt_state.items()):
        if src_key != tgt_key:
            raise ValueError(
                f"Expected source and target state dict to match ({src_key} != {tgt_key})"
            )
        if src_key not in map_extend_embeddings:
            # ignore this key
            continue

        if src_key.endswith("weight") and src_tensor.size(1) != tgt_tensor.size(1):
            raise ValueError(
                "Mapping to a smaller embedding dimension is not supported",
            )

        if (src_n_ids := src_tensor.size(0)) > tgt_tensor.size(0):
            raise ValueError(
                "Mapping to a smaller number of embeddings is not supported",
            )
        tgt_state[tgt_key][:src_n_ids] = src_tensor.detach()
        notify(f"Successfully moved embeddings from {src_key}")
    return tgt_state


def _modules_map_rename(
    src_state: dict[str, torch.Tensor],
    map_rename_modules: Iterable[tuple[str, str]],
    notify: Callable[[str], Any],
) -> dict[str, torch.Tensor]:
    """
    Rename entries in a state dict in place by providing a map of
    (source_prefix, target_prefix). As soon as a compatible entry is found,
    i.e., a key in the state dict matches `source_prefix`, `source_prefix` is
    replaced with `target_prefix`
    """
    src_state_updated = src_state.copy()
    for rename_from_refix, rename_to_prefix in map_rename_modules:
        for src_module_name in src_state.keys():
            if src_module_name.startswith(rename_from_refix):
                data = src_state_updated.pop(src_module_name)
                postfix = src_module_name[len(rename_from_refix) :]
                tgt_module_name = rename_to_prefix + postfix
                src_state_updated[tgt_module_name] = data
                notify(f"Successfully renamed {src_module_name +postfix } to {tgt_module_name}")
    return src_state_updated


@torch.no_grad()
def load_model_state(
    checkpoint_file: str | Path,
    model: _TModule,
    map_location: MAP_LOCATION | None = None,
    map_extend_embeddings: set[str] | None = None,
    map_rename_modules: Iterable[tuple[str, str]] | None = None,
    on_success: Callable[[str], Any] | None = None,
) -> tuple[_TModule, float]:
    """
    Restore a model's `state_dict` from a checkpoint and return it.

    Args:
        checkpoint_file: Must point to a `.pth` file
        model: Any class T that inherits from `torch.nn.Module`
        map_location: See `torch.load`
        map_extend_embeddings: Map a smaller source embedding to a
            larger target embedding
        map_rename_modules: Rename modules in the checkpoint state
            before populating the model with it. If provided in the form
            (source_prefix, target_prefix), the state dict will be
            renamed BEFORE `map_extend_embeddings` is applied!

    Returns:
        tuple: A tuple with the original model `T` with
            `updated state_dict` and the associated loss
    """
    snapshot = torch_load(
        checkpoint_file,
        map_location,
        weights_only=True,
    )
    state_dict = snapshot["MODEL"]
    loss = snapshot["LOSS"]

    def _notify(msg: str):
        if on_success:
            on_success(msg)

    if map_rename_modules:
        state_dict = _modules_map_rename(state_dict, map_rename_modules, notify=_notify)
    if map_extend_embeddings:
        state_dict = _embedding_map_extend(
            state_dict,
            model.state_dict(),
            map_extend_embeddings,
            notify=_notify,
        )
    model.load_state_dict(state_dict, strict=True)
    return model, loss



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from dataclasses import dataclass
from typing import Any, Callable, Generator, Generic, Iterator, Protocol, TypeVar

_T = TypeVar("_T", covariant=True)


class _SizedIterable(Protocol[_T]):
    """
    An internal protocol type used by the epoch_cycler, typically to allow Torch
    DataLoaders as iterables for the cycler. Ideally, we'd have the input
    sequence be `Sequence` to facilitate the logic in the epoch cycler. However,
    Dataloaders are not `Sequence` because they do not implement `__getitem__`
    and thus cannot be used to efficiently index. But since DataLoaders are both
    `Sized` and `Iterator`, we can make use that. See
    https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes
    """

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[_T]: ...


# TODO: Python 3.12, NamedTuple
@dataclass
class EpochCyclerYield(Generic[_T]):
    epoch: int
    batch: int
    item: _T
    next_epoch: int
    next_batch: int


def epoch_cycler(
    seq: _SizedIterable[_T],
    *,
    before_new_epoch: Callable[[int], Any] | None = None,
    start_epoch: int = 0,
    start_batch: int = 0,
    max_iters: int | None = None,
) -> Generator[EpochCyclerYield[_T], None, None]:
    """
    Infintely iterates over a sequence, yielding batches and their indices as
    well as the current epoch (one complete iteration over the sequence).

    Args:
        seq (_SizedIterable[T]): An object that implements both __size__ and
            __len__ holding items T
        before_new_epoch (Callable[[int], Any] | None = None): Optional
            callback function that is called right **before** the `n`-th starts
            with `n`. Can be used to shuffle a dataset or perform arbitry side effects
        start_epoch (int = 0): Start/resume from this epoch
        start_batch (int = 0): Skip until and yield **starting from** this index
        max_iters (int | None = None): Return when this total amount of batches has
            been yielded across epochs

    Returns:
        Generator (EpochCyclerYield): Yields the epoch, the batch index and the item at
            that index. Batch indices are reset at the start of each epoch (that is,
            they range from 0 to len(seq)). In order to restore the iterator's state,
            the next epoch and next batch index are also returned for convenience.
    """
    epoch = start_epoch
    global_batch_counter = 0
    if start_batch >= len(seq):
        raise IndexError("start_batch is larger than the length of the sequence")

    if before_new_epoch:
        before_new_epoch(epoch)

    it = iter(seq)
    # advance the first iterator to the start batch index
    for _ in range(0, start_batch):
        next(it)

    while True:
        # do not factor out the call to len - if the sequence changes between
        # epochs, we want to account for that. len is O(1), doesn't matter
        for i in range(start_batch, len(seq)):
            if global_batch_counter == max_iters:
                return
            item = next(it)
            if i == len(seq) - 1:
                next_epoch = epoch + 1
                next_batch = 0
            else:
                next_epoch = epoch
                next_batch = i + 1
            yield EpochCyclerYield[_T](epoch, i, item, next_epoch, next_batch)
            global_batch_counter += 1
        start_batch = 0
        epoch += 1

        if before_new_epoch:
            before_new_epoch(epoch)
        it = iter(seq)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

"""
This module defines the core attributes that need to be specified for an
experiment. More granular training runs may subclass the core classes and add
additional attributes.
"""

from typing import Generic, Literal, NamedTuple, TypeVar

from pydantic import BaseModel, ConfigDict, Field


class CoreModelParams(BaseModel):
    """
    The core model parameters that all models must specify when training with
    the `CoreTrainer`.
    """

    input_seq_len: int = Field(
        description="The input sequence length. This can be regarded as the context size when training for language. In other scenarios like linear regression, this might be a single data point, hence 1"
    )


class CoreTrainConfig(BaseModel):
    """
    The core training parameters needed for a training run with the
    `CoreTrainer`.
    """

    target_elements: int = Field(
        description="The desired number of data points to train on. If `None`, defaults to using data points from the training set once, resulting in a single epoch. Note that this is a *lower bound* - due to the sequence length and batch sizes, in effect, we train on more than this target. Use this when you want to train on a fixed subset of data, e.g., a number of bytes"
    )
    target_elements_strategy: Literal["batch", "sequence"] = Field(
        description="The strategy to count elements in a batch, used to determine when `target_elements` is achieved. 'batch' means batch size. 'sequence' will count each element in the batch as contributing to the `target_elements`. E.g., when a batch has `n` sequences of length `L`, then the 'sequence' strategy will count `n` * `L` target elements per batch"
    )
    warmup_steps_perc: float = Field(
        default=0.1,
        description="A float in the range [0, 1] to determine how many of the total gradient steps should be used for a warmup. The rest of the steps follows cosine annealing",
    )
    batch_size: int = Field(description="The batch size")

    shuffle_train: bool = Field(
        default=False,
        description="Shuffle the training data in the data loader. Enable this only if you are sure you're not chaining runs",
    )
    shuffle_eval: bool = Field(
        default=False, description="Shuffle the validation data in the data loader"
    )
    max_eval_steps: int | None = Field(
        default=None, description="Maximum number of evaluation iterations"
    )
    learning_rate: float = Field(description="The learning rate")
    gradient_clipping: float | None = Field(
        default=None, description="A value for gradient clipping"
    )
    gradient_accumulate_every: int = Field(
        description="After how many batches to accumulate the gradient"
    )


class ResumeConfig(BaseModel):
    """
    The resume parameters needed for to resume training from a checkpoint and a
    specific epoch and batch with `CoreTrainer`. In case training starts for the
    first time, this class is not needed. For a completed training run, it is
    automatically populated so that it can be used in a future training run.
    """

    checkpoint_file: str = Field(description="Path to model state checkpoint")
    next_epoch_index: int = Field(description="The epoch to resume from")
    next_batch_index: int = Field(description="The global batch counter to resume from")
    migrate_embeddings: bool = Field(
        default=False, description="Migrate an existing smaller number of embeddings"
    )
    rename_modules: bool = Field(
        default=False,
        description="Rename modules from existing models to new names",
    )
    resumed_from: str | None = Field(
        default=None,
        description="If training has been resumed from a previous checkpoint/experiment, points to the config of that experiment to easily trace chained training runs",
    )


class CoreIoConfig(BaseModel):
    """
    The core input/output parameters needed for a training run with the
    `CoreTrainer`.
    """

    name_model: str = Field(description="Model name for saving checkpoints")
    output_dir: str = Field(
        description="The output directory for all artefacts (will be created automatically based on `model_name` and a unique postfix)"
    )
    num_models_to_save: int = Field(
        description="Max number of best performing models to store. If smaller than `validate_amount`, stores only `validate_amount` models"
    )
    validate_amount: int = Field(
        description="How often (in total) to run the validation set and reserve a model candidate. **Must be >= than `num_models_to_save`**"
    )
    log_train_loss_amount: int = Field(description="How often (in total) to log training loss")
    enabled_loss_log_for_gpus: list[int] = Field(
        default=[0], description="The rank of the GPUs that should write to the CSV loss file"
    )


TModelParams = TypeVar("TModelParams", covariant=True, bound=CoreModelParams)
TTrainConfig = TypeVar("TTrainConfig", covariant=True, bound=CoreTrainConfig)
TIoConfig = TypeVar("TIoConfig", covariant=True, bound=CoreIoConfig)


class SummaryStats(BaseModel):
    """
    Summary stats for a training run
    """

    training_start: str
    training_end: str
    num_workers: int
    cuda_devices: list[str]
    parameter_count: int
    error: str | None


class GenericEntryConfig(BaseModel, Generic[TModelParams, TTrainConfig, TIoConfig]):
    """
    A generic entry config for training with the trainer that can be made more
    specific by subclassing.
    """

    model_config = ConfigDict(strict=True)

    params: TModelParams
    train: TTrainConfig
    io: TIoConfig
    resume: ResumeConfig | None = None


class GenericOutputConfig(BaseModel, Generic[TModelParams, TTrainConfig, TIoConfig]):
    """
    A generic output config for a training run with the trainer that can be made
    more specific by subclassing.
    """

    model_config = ConfigDict(strict=True)

    params: TModelParams
    train: TTrainConfig
    io: TIoConfig
    resume: ResumeConfig
    summary: SummaryStats


class CSVLossEntry(NamedTuple):
    gpu_rank: int
    timestamp: str
    elements_seen: int
    kind: str
    epoch: int
    batch: int
    cum_batch: int
    loss: float
    lr: float
    avg_grad: float
    avg_grad_clipped: float


class CSVTimeAndMemSnapshotEntry(NamedTuple):
    cum_batch: int
    num_items: int
    kind: str
    fw_time: float
    bw_time: float | None
    allocated: float
    allocated_max: float
    reserved: float
    reserved_max: float
    total: float



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from dataclasses import dataclass
from functools import partial
from typing import Literal

import torch
import torch.version
from torch.types import Device


@dataclass
class CudaProperties:
    cuda_available: bool
    cuda_version: str | None
    cuda_devices: list[str]


def cuda_properties() -> CudaProperties:
    is_cuda_available = torch.cuda.is_available()
    devices: list[torch._C._CudaDeviceProperties] = [
        torch.cuda.get_device_properties(i) for i in range(torch.cuda.device_count())
    ]
    return CudaProperties(
        cuda_available=is_cuda_available,
        cuda_version=torch.version.cuda,
        cuda_devices=[f"{d.name}_{d.total_memory}_{d.major}.{d.minor}" for d in devices],
    )


@dataclass
class CudaMemorySnapshot:
    allocated: float
    allocated_max: float
    reserved: float
    reserved_max: float
    total: float
    unit: Literal["GB"] = "GB"


_gb_conv_factor = 1024**3


def cuda_memory_snapshot(device: Device | None, dec_prec: int = 3) -> CudaMemorySnapshot:
    if not torch.cuda.is_available():
        return CudaMemorySnapshot(
            allocated=0.0,
            allocated_max=0.0,
            reserved=0.0,
            reserved_max=0.0,
            total=0.0,
        )
    # report only in gb for now
    _round = partial(round, ndigits=dec_prec)
    alloc = _round(torch.cuda.memory_allocated(device) / _gb_conv_factor)
    max_alloc = _round(torch.cuda.max_memory_allocated(device) / _gb_conv_factor)
    res = _round(torch.cuda.memory_reserved(device) / _gb_conv_factor)
    max_res = _round(torch.cuda.max_memory_reserved(device) / _gb_conv_factor)
    current_device: torch._C._CudaDeviceProperties = torch.cuda.get_device_properties(device)
    total = _round(current_device.total_memory / _gb_conv_factor)
    return CudaMemorySnapshot(alloc, max_alloc, res, max_res, total)



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

from typing import Iterable

import torch


def target_loss_mask(
    tensor_weight_pairs: Iterable[tuple[torch.Tensor, float]],
) -> torch.Tensor:
    """
    Create a loss mask for an iterable of pairs of tensor, mask_weight.
    """
    loss_mask_tensors: list[torch.Tensor] = []
    for tensor, loss_weight in tensor_weight_pairs:
        loss_mask_tensors.append(torch.full_like(tensor, dtype=torch.float, fill_value=loss_weight))
    return torch.concat(loss_mask_tensors)


def shift_remap_tensor(
    tensor: torch.Tensor, range_start: int
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Remaps the values of a tensor to a new range starting from `range_start`.
    Returns the remapped tensor of the same shape and dtype and two tensors to
    map back to the original.

    Example:

        shifted, unshift, indices = remap_values(input_tensor, range_start=11)
        assert input_tensor.equal(unshift[indices])
    """
    uniqued: tuple[torch.Tensor, torch.Tensor] = torch.unique(
        tensor, return_inverse=True, sorted=True
    )
    # inverse_indices is the indices of the original elements in the returned
    # unique_values tensor
    unique_values, inverse_indices = uniqued
    remapped_values = torch.arange(
        range_start, range_start + unique_values.size(0), dtype=tensor.dtype
    )
    remapped_tensor = remapped_values[inverse_indices].reshape_as(tensor)

    return remapped_tensor, unique_values, inverse_indices



from __future__ import annotations

__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import os
import random
from pathlib import Path
from typing import TYPE_CHECKING, Generator

import torch
import tqdm
from pydantic import BaseModel
from typing_extensions import Unpack

from mblm.data.datasets import DistributedDataset, DistributedDatasetConfig
from mblm.data.types import BatchWithLossMask, ModelMode
from mblm.data.utils import Bytes

if TYPE_CHECKING:
    from mblm.train.mblm import TrainEntryConfig


class PG19ModelGeneration(BaseModel):
    id_model: str
    book_id: str
    book_txt_offset: int
    ctx_len: int
    generated: list[int]
    truth: list[int]
    ce: float
    generation_time: float
    timestamp: str


class PG19(DistributedDataset[BatchWithLossMask]):
    """
    https://github.com/google-deepmind/pg19

    The `data_dir` is expected to be the of the exact structure as the
    original dataset, although only the test, train and validation folders
    are strictly needed:
        ├── LICENSE
        ├── README.md
        ├── metadata.csv
        ├── test
        ├── train
        └── validation

    """

    def __init__(
        self,
        data_dir: str | Path,
        mode: ModelMode,
        load_mininterval: int = 30,
        display_load_progress: bool = True,
        **config: Unpack[DistributedDatasetConfig],
    ):
        root = Path(data_dir)
        if mode == ModelMode.VALID:
            data_path = root / "validation"
        else:
            data_path = root / mode.value
        self.txt_files = [data_path / file for file in os.listdir(data_path)]

        data_buff = bytearray()
        for file in tqdm.tqdm(
            self.txt_files,
            desc=f"Loading pg19 {data_path}",
            mininterval=load_mininterval,
            disable=not display_load_progress,
        ):
            with Path.open(file, "rb") as f:
                data_buff.extend(f.read())
        self.data = Bytes.bytes_to_tensor(data_buff)
        super().__init__(
            data_size=self.data.numel(),
            is_sequential=True,
            **config,
        )

    @staticmethod
    def from_train_entry_config(
        config: TrainEntryConfig,
        mode: ModelMode,
        worker_id: int,
        num_workers: int,
    ) -> DistributedDataset[BatchWithLossMask]:
        return PG19(
            data_dir=config.io.dataset_dir,
            mode=mode,
            seq_len=config.params.input_seq_len,
            worker_id=worker_id,
            num_workers=num_workers,
        )

    @staticmethod
    def supports_test_mode() -> bool:
        return True

    def get_sample(self, from_idx: int) -> BatchWithLossMask:
        """
        Get a sample with a loss mask. This method is required by the
        DistributedDataset superclass.
        """
        sample = self.data[from_idx : from_idx + self.seq_len].long()
        return sample, torch.ones_like(sample, dtype=torch.float16)

    def book(self, name: str) -> str:
        """
        Get a book by its name (e.g., `44381.txt`) return its content as a
        string
        """
        for candidate in self.txt_files:
            if candidate.name == name:
                with Path.open(candidate, "r", encoding="utf8") as f:
                    return f.read()
        raise ValueError(f"Book {name} does not exist")

    def iter_sequences_rand(self) -> Generator[torch.Tensor, None, None]:
        """
        Iterate over random sequences across books of PG19
        """
        max_sample_start_idx = len(self.data) - self.seq_len - 1
        while True:
            idx = random.randint(0, max_sample_start_idx)
            yield self.data[idx : idx + self.seq_len]

    def iter_books(self, shuffle: bool = False) -> Generator[tuple[str, str], None, None]:
        """
        Iterate over all the books in PG19, possibly in random order. Return an
        iterator over the index of the book and its content as a string
        """
        txt_file_idxs = list(range(len(self.txt_files)))
        if shuffle:
            random.shuffle(txt_file_idxs)
        for i in txt_file_idxs:
            book = self.txt_files[i]
            with Path.open(book, "r", encoding="utf8") as f:
                yield book.name, f.read()



__copyright__ = """MIT License

Copyright (c) 2024 - IBM Research

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import copy
import heapq
from dataclasses import dataclass, field
from typing import Generic, Iterator, SupportsFloat, TypeVar

_T = TypeVar("_T")


@dataclass(order=True)
class _Entry(Generic[_T]):
    val: float
    item: _T = field(compare=False)


class TopN(Generic[_T]):
    """
    A priority queue based on a heap data structure. By default, a min heap is created,
    with the first element being the smallest. A max heap can be specified via `top_largest`.

    Args:
        n (int): Max number of items to store
        deep_copy (bool = `False`): Create a deep copy of elements
        top_largest (bool = `False`): If true, store the `n` largest items instead of
            the smallest items

    Returns:
        self (TopN[T]): An iterable instance of `self` that returns the smallest items
            (or largest of `top_largest` is set to `True`) first

    **Example**::

        # create the heap
        top_n = TopN[str](2)

        # add items in random order
        top_n.add((1, "a"))
        top_n.add((3, "b"))
        top_n.add((2, "c"))

        for idx, item in enumerate(top_n):
            ...
            # iteration 1: (1, "a")
            # iteration 2: (2, "c")
    """

    def __init__(self, n: int, *, deep_copy: bool = False, top_largest: bool = False):
        self._max_heap_items = n
        self._heap: list[_Entry] = []
        self._top_largest = top_largest
        self._heap_factor = 1.0 if top_largest else -1.0
        self._deep_copy = deep_copy

    def add(self, item: tuple[SupportsFloat, _T]) -> None:
        """
        Add an item to the queue. The first tuple entry is used to
        determine the position of the newly added element
        """
        val, data = item
        if self._deep_copy:
            data = copy.deepcopy(data)
        entry = _Entry(self._heap_factor * float(val), data)
        curr_heap_items = len(self._heap)

        # heap is underfull
        if curr_heap_items < self._max_heap_items:
            heapq.heappush(self._heap, entry)
        # heap is full, maybe add
        elif entry > self._heap[0]:
            heapq.heapreplace(self._heap, entry)

    def get_top(self, best: int | None = None) -> list[tuple[float, _T]]:
        """
        Get the n `best` (i.e., smallest or largest depending on initialization)
        items from the queue. If `best` is not specified, all items are returned
        """
        if not best:
            best = self._max_heap_items
        top_n = sorted(
            [(self._heap_factor * x.val, x.item) for x in self._heap],
            key=lambda x: x[0],
            reverse=self._top_largest,
        )
        return top_n[:best]

    def __iter__(self) -> Iterator[tuple[float, _T]]:
        return iter(self.get_top())

    def __len__(self) -> int:
        return len(self._heap)
